import pathlib
import networkx as nx
from ast import literal_eval
from numpy import genfromtxt, minimum
from scipy.spatial.distance import cdist
import matplotlib.pyplot as plt
from matplotlib.colors import LinearSegmentedColormap
from statistics import mean
import warnings

# Special characters for parsing files
COMMENT_CHAR = '#'
DEFAULT_CHAR = '.'
BASKET_LABEL = '@'

DISTANCE_SCALE_FACTOR = 32*4

# Constants for plotting graphs
EXTENTS = {'ap': (660, 300, 760),
           'lo': (396, 222, 1074),
           'mb': (570, 260, 870),
           'sg': (888, 640, 850),
           'ss': (363, 526, 720)}

# Colormap for plotting strategies
HIGH_CDICT = {'red':   [[0.0, 1.0, 1.0],
                        [1.0, 1.0, 1.0]],
              'green': [[0.0, 0.8, 0.8],
                        [0.6, 0.8, 0.8],
                        [1.0, 1.0, 1.0]],
              'blue':  [[0.0, 0.0, 0.0],
                        [0.6, 0.0, 0.0],
                        [1.0, 0.6, 0.6]]}

LOW_CDICT =  {'red':   [[0.0, 0.0, 0.0],
                        [0.6, 0.0, 0.0],
                        [1.0, 0.6, 0.6]],
              'green': [[0.0, 0.6, 0.6],
                        [1.0, 0.9, 0.9]],
              'blue':  [[0.0, 0.3, 0.3],
                        [0.6, 0.4, 0.4],
                        [1.0, 0.7, 0.7]]}

high_cmap = LinearSegmentedColormap('HighPath', segmentdata=HIGH_CDICT, N=256)
low_cmap = LinearSegmentedColormap('LowPath', segmentdata=LOW_CDICT, N=256)


# noinspection PyTypeChecker,PyTypeChecker
class GusherMap:
    def __init__(self, map_id, weights=None, squad=False):
        self.map_id = map_id
        self._path = pathlib.Path(__file__).parent.resolve() / f'maps/{map_id}'

        self._load_gushers(str(self._path/'gushers.csv'))
        self._load_distances(str(self._path/'distance_modifiers.txt'), squad)
        self._validate_distances()
        self._load_connections(str(self._path/'connections.txt'))
        if not weights:
            # Read the weight dictionary from the first non-commented line of the file
            # https://stackoverflow.com/a/26284995
            f = (line for line in open(str(self._path/'weights.txt'))
                 if not line.lstrip().startswith(COMMENT_CHAR))
            weights = next(f).strip()
        self._load_weights(literal_eval(weights))

    def _load_gushers(self, filename):
        self._gushers = genfromtxt(filename, delimiter=',', names=['name', 'coord'], dtype=['U8', '2u4'])

    # TODO - use tkg's exact distances
    def _load_distances(self, filename, squad=False):
        coords = self._gushers['coord']
        # Read norm from the 2nd line of the file
        with open(filename) as f:
            f.readline()
            norm_raw = f.readline().split(': ')[-1].strip(' \n')
        norm = float(norm_raw)
        adjacency_matrix = cdist(coords, coords, 'minkowski', p=norm) / DISTANCE_SCALE_FACTOR
        try:
            distance_modifiers = genfromtxt(filename, delimiter=',', comments=COMMENT_CHAR)
            adjacency_matrix += distance_modifiers
            if squad:
                # Assume that it never takes longer to reach a gusher than it would have taken coming from basket/spawn
                adjacency_matrix = minimum(adjacency_matrix, adjacency_matrix[0, :])
        except ValueError as e:
            warnings.warn(f"Couldn't read distance modifiers from '{filename}'\n" + str(e))
        self.distances = nx.from_numpy_array(adjacency_matrix, create_using=nx.DiGraph)
        # noinspection PyTypeChecker
        nx.relabel_nodes(self.distances, lambda i: self._gushers['name'][i], False)

    # May not be necessary
    def _load_distances_all_equal(self, nodes, all_distances=1, norm=2):
        self.distances = nx.complete_graph(nodes, create_using=nx.DiGraph)
        nx.set_edge_attributes(self.distances, all_distances, name='weight')
        # Use real distances for outgoing edges from the basket
        basket_distances = cdist(self._gushers['coord'][0].reshape(1, 2),
                                 self._gushers['coord'][1:],
                                 metric='minkowski', p=norm) / DISTANCE_SCALE_FACTOR
        edge_list = [(BASKET_LABEL, self._gushers['name'][i+1], basket_distances[0][i]) for i in range(len(nodes))]
        self.distances.add_node(BASKET_LABEL)
        self.distances.add_weighted_edges_from(edge_list)

    def _validate_distances(self):
        violations = self._find_triangle_inequality_violations()
        if violations:
            warnings.warn(f"Distances matrix for map '{self.map_id}' does not satisfy triangle inequality:\n" +
                          ''.join(f"    {t[0]}->{t[1]}->{t[2]} ({t[3]:g}) is shorter than {t[0]}->{t[2]} ({t[4]}:g)\n"
                                  for t in violations))

    def _load_connections(self, filename):
        # Read the map name from the first line of the file
        with open(filename) as f:
            self.name = f.readline().strip(COMMENT_CHAR + ' \n')
        connections_raw = nx.read_adjlist(filename, comments=COMMENT_CHAR)
        conn_size = len(connections_raw)
        dist_size = len(self.distances)
        assert dist_size == conn_size + 1, f"Couldn't read {filename}\n" + \
                                           f"Distances matrix is {dist_size}x{dist_size} " + \
                                           f"but connections graph has {conn_size} vertices"
        self.connections = self.distances.edge_subgraph(connections_raw.to_directed(as_view=True).edges)
        # There's probably some way to induce a subgraph of 'distances' by reading edges directly from connections.txt,
        #   but I'm too lazy to figure out what that is

    def _load_weights(self, weights_dict):
        self.weights = {BASKET_LABEL: 0}
        for gusher in self.connections:
            gusher_weight = weights_dict[DEFAULT_CHAR]
            for group in weights_dict:
                if gusher in group:
                    gusher_weight = weights_dict[group]
                    break
            self.weights[gusher] = gusher_weight

    def _find_triangle_inequality_violations(self):
        def distance(start, end):
            return self.distances.adj[start][end]['weight']
        violations = set()
        for vertex in self.distances:
            neighborhood = set(self.distances.adj[vertex])
            for neighbor in neighborhood:
                shortest_distance = distance(vertex, neighbor)
                for other in neighborhood.difference({neighbor}):
                    other_distance = distance(vertex, other) + distance(other, neighbor)
                    if other_distance < shortest_distance:
                        violations.add((vertex, other, neighbor, other_distance, shortest_distance))
        return violations

    def __len__(self):
        return len(self.connections)

    def __iter__(self):
        return self.connections.__iter__()

    def __contains__(self, item):
        return item in self.connections

    def distance(self, start, end):
        """Return distance between two gushers."""
        return self.distances.adj[start][end]['weight']

    def weight(self, vertex):
        """Return the weight of a gusher."""
        return self.weights[vertex]

    def adj(self, vertex):
        """Return adjacent gushers for a given gusher."""
        return self.connections.adj[vertex]

    def degree(self, vertex):
        """Return the number of adjacent gushers for a given gusher."""
        return self.connections.degree[vertex]

    def plot(self, strategy=None, tuning=0.5):
        background = plt.imread(str(self._path.parent.parent.resolve()/f'images/{self.map_id}.png'))
        pos = {gusher['name']: tuple(gusher['coord']) for gusher in self._gushers if gusher['name'] != BASKET_LABEL}
        pos_attrs = {node: (coord[0] - 40, coord[1]) for (node, coord) in pos.items()}

        if self.map_id in EXTENTS:
            x, y, length = EXTENTS[self.map_id]
            extent = [x, x+length, y+length, y]
            background = background[y:y+length, x:x+length]
        else:
            extent = None

        fig = plt.figure(figsize=(6, 6))
        ax = plt.subplot()
        ax.set_facecolor('#444444')
        plt.imshow(background, extent=extent)
        plt.title(self.name)
        nx.draw_networkx_edges(nx.to_undirected(self.connections), pos,
                               edge_color='#888888', style='dashed', width=1.5)
        nx.draw_networkx_labels(self.connections, pos_attrs, labels={gusher: self.weight(gusher) for gusher in pos},
                                font_weight='bold', font_color='#ff4a4a', horizontalalignment='right')

        if strategy:
            latencies, risks = strategy.get_costs(self)
            key = f"average time: {mean(latencies.values()):0.2f}, worst time: {max(latencies.values()):0.2f}\n" \
                  f"average risk: {mean(risks.values()):0.2f}, worst risk: {max(risks.values()):0.2f}"
            if tuning is not None:
                key = f"setting: {100*(1 - tuning):g}% speed, {100*tuning:g}% safety\n" + key
            # https://stackoverflow.com/a/32745842
            ax.annotate(key, xy=(0, 0), xytext=(0, 10),
                        xycoords=('axes fraction', 'figure fraction'),
                        textcoords='offset points',
                        size=10, ha='left', va='bottom')

            nonfindable = tuple(strategy.nonfindable_nodes())
            nonfindable_str = tuple(str(node) for node in nonfindable)
            if nonfindable:
                pos.update({str(node): (pos[node.name][0] + 50, pos[node.name][1]) for node in nonfindable})
            strat_graph = nx.to_networkx_graph(strategy.get_adj_dict(), create_using=nx.DiGraph)

            def node_color(node):
                if node == strategy.name:
                    return '#2bff2b'
                elif node in nonfindable_str:
                    return '#aaaaaa'
                else:
                    return '#ffffff'
            node_colors = [node_color(node) for node in strat_graph.nodes]
            high_edges = [(s, t) for s, t in strat_graph.edges if strat_graph[s][t]['high']]
            high_colors = [strat_graph[s][t]['depth'] for s, t in high_edges]
            low_edges = [(s, t) for s, t in strat_graph.edges if not strat_graph[s][t]['high']]
            low_colors = [strat_graph[s][t]['depth'] for s, t in low_edges]
            color_kwargs = ({'edgelist':  high_edges, 'edge_color': high_colors, 'edge_cmap': high_cmap},
                            {'edgelist': low_edges, 'edge_color': low_colors, 'edge_cmap': low_cmap})
            for kwargs in color_kwargs:
                nx.draw_networkx_edges(strat_graph, pos,
                                       width=2, connectionstyle='arc3, rad=0.25', min_target_margin=12,
                                       arrowstyle='simple, head_length=1.2, head_width=1.2', **kwargs)

            nx.draw_networkx_nodes(strat_graph, pos, node_color=node_colors)
        else:
            nx.draw_networkx_nodes(self.connections, pos, node_color='#ffffff')

        nx.draw_networkx_labels(self.connections, pos, labels={gusher: gusher for gusher in pos}, font_color='#111111')

        plt.show()


# TODO - move to separate test file
if __name__ == '__main__':
    for map_id in ('sg', 'ss', 'mb', 'lo', 'ap'):
        gusher_map = GusherMap(map_id)
        print('-'*len(gusher_map.name))
        print(gusher_map.name)
        for node in gusher_map:
            print(f"gusher {node} (weight {gusher_map.weight(node):g}) is adjacent to " +
                  ', '.join(f"{v} ({e['weight']:0.2f})" for v, e in gusher_map.adj(node).items()))
        gusher_map.plot()



import click
import pathlib
from os import scandir
from . import __version__
from .GusherMap import GusherMap
from .GusherNode import read_tree
from .strats import get_strat


# TODO - start compilation of strategy variants for each gushers

# Settings for Click
CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])

# Get list of available map IDs
HERE = pathlib.Path(__file__).parent.resolve()
maps = [f.name for f in scandir(HERE/'maps/') if f.is_dir()]


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('--map', '-m', 'map_id', required=True,
              type=click.Choice(maps, case_sensitive=False),
              help="""Map ID. Must be the name of a folder in 'goldieseeker/maps'.""")
@click.option('--tuning', '-t', type=click.FloatRange(0, 1), default=0.5,
              help="""\b
              Set the seeking algorithm's tuning factor (range between 0-1).
              0: fastest time, 1: lowest risk""")
@click.option('--squad', '-s', is_flag=True,
              help="""\b
              Turn on "squad" mode (experimental).
              This tells the seeking algorithm to assume that traveling to a gusher never takes longer than it would 
              take when coming from spawn (i.e. the basket).""")
@click.option('--eval', '-E', 'strategy_str', type=str,
              help="""\b
              Evaluate a user-specified strategy.
              example: -E "f(d(b, g), e(c, a))"
              "a(b, c)" means "if A is high, open B, and if A is low, open C\"""")
@click.option('--weights', '-W', type=str,
              help="""\b
              Specify custom gusher weights in dictionary format.
              example: -W "{'d': 4, 'bef': 2, '.': 1}"
              This gives a weight of 4 to gusher D, a weight of 2 to gushers B, E, and F, """
              "and a weight of 1 to the rest.")
@click.option('--quiet', '-q', count=True,
              help="""\b
              Don't show the map plot.
              Use '-qq' to also suppress reporting strategy details.
              Use '-qqq' to only output the string representation of the strategy tree.""")
@click.option('--debug', '-d', is_flag=True,
              help="Print internal process of search algorithm.")
@click.version_option(__version__, '--version', '-v', prog_name="goldieseeker")
def main(map_id, tuning, squad, strategy_str, weights, quiet, debug):
    """\b
    For a given map, generate a Goldie Seeking strategy or evaluate a user-specified strategy.
    To customize default distances and weights, edit the corresponding files in goldieseeker/maps/[MAP_ID]."""
    try:
        gusher_map = GusherMap(map_id, weights=weights, squad=squad)
    except IOError as err:
        click.echo(f"Couldn't load map '{map_id}'!", err=True)
        click.echo(str(err), err=True)
    else:
        if strategy_str:
            strat = read_tree(strategy_str, gusher_map)
            strat.validate(gusher_map)
        else:
            strat = get_strat(gusher_map, tuning=tuning, debug=debug)
            # strat.validate(gusher_map)
        click.echo(strat.report(gusher_map, quiet=quiet))
        if quiet < 1:
            gusher_map.plot(strat, tuning if not strategy_str else None)


if __name__ == '__main__':
    main()



from .GusherMap import BASKET_LABEL
from copy import deepcopy
from statistics import mean
from statistics import pstdev
from pyparsing import Regex, Forward, Suppress, Optional, Group

# Flag to indicate gusher is non-findable
NEVER_FIND_FLAG = '*'


# TODO - switch to using anytree
class GusherNode:
    def __init__(self, name, gusher_map=None, findable=True):
        self.name = name
        self.low = None  # next gusher to open if this gusher is low
        self.high = None  # next gusher to open if this gusher is high
        self.parent = None  # gusher previously opened in sequence
        self.findable = findable  # whether it is possible to find the Goldie at this gusher
        # if findable is False, the gusher is being opened solely for information (e.g. gusher G on Ark Polaris)
        # non-findable nodes still count towards their children's costs, but don't count towards tree's objective score
        self.size = 1 if findable else 0  # number of findable nodes in subtree rooted at this node
        self.distance = 1  # distance from parent gusher
        self.latency = 0  # if Goldie is in this gusher, how long it takes to find Goldie by following decision tree
        # latency = total distance traveled on the path from root node to this node
        self.total_latency = 0  # sum of latencies of this node's findable descendants
        if gusher_map:
            self.weight = gusher_map.weight(name)  # risk weight for this gusher
        else:
            self.weight = 1
        self.risk = 0  # if Goldie is in this gusher, roughly how much trash is spawned by following decision tree
        # Trash spawned by a given gusher is multiplied by the gusher's weight
        # This does not mean the gusher actually spawns more fish; it is just a way of telling the algorithm that
        #   some gushers spawn more dangerous trash than others (e.g. gushers next to basket)
        self.total_risk = 0  # sum of risks of this node's findable descendants

    def __str__(self):
        return self.name + (NEVER_FIND_FLAG if not self.findable else "")

    def __repr__(self):
        parent = str(self.parent) if self.parent else BASKET_LABEL
        repr_str = f'{parent}-->{self}'
        if self.high and self.low:
            repr_str += f'({self.high}, {self.low})'
        elif self.high:
            repr_str += f'({self.high},)'
        elif self.low:
            repr_str += f'({self.low},)'
        return repr_str
        # return write_tree(self)  + f'; time: {self.total_latency}, risk: {self.total_risk}}}'

    def __iter__(self):
        yield self
        if self.high:
            yield from self.high.__iter__()
        if self.low:
            yield from self.low.__iter__()

    def __eq__(self, other):
        return isinstance(other, GusherNode) and write_tree(self) == write_tree(other)

    # Override deepcopy so that it does not copy non-root nodes' cost attributes (weight, size, latency, etc.)
    # This improves performance without sacrificing any accuracy
    # noinspection PyDefaultArgument
    def __deepcopy__(self, memodict={}):
        tree_copy = GusherNode(self.name, findable=self.findable)
        if not self.parent:
            cost_attrs = ('size', 'distance', 'latency', 'total_latency', 'weight', 'risk', 'total_risk')
            tree_copy.__dict__.update({attr: self.__dict__.get(attr) for attr in cost_attrs})
        if self.high:
            tree_copy.high = deepcopy(self.high)
            tree_copy.high.parent = tree_copy
        if self.low:
            tree_copy.low = deepcopy(self.low)
            tree_copy.low.parent = tree_copy
        return tree_copy

    def add_children(self, high, low, dist_h=1, dist_l=1):
        size_h, size_l = 0, 0
        totlat_h, totlat_l = 0, 0
        totrisk_h, totrisk_l = 0, 0
        if high:
            assert not self.high, f'gusher {self} already has high child {self.high}'
            assert not high.parent, f'gusher {high} already has parent {high.parent}'
            self.high = high
            self.high.parent = self
            self.high.distance = dist_h
            size_h = self.high.size
            totlat_h = self.high.total_latency
            totrisk_h = self.high.total_risk
        if low:
            assert not self.low, f'gusher {self} already has low child {self.low}'
            assert not low.parent, f'gusher {low} already has parent {low.parent}'
            self.low = low
            self.low.parent = self
            self.low.distance = dist_l
            size_l = self.low.size
            totlat_l = self.low.total_latency
            totrisk_l = self.low.total_risk
        self.size = size_l + size_h + (1 if self.findable else 0)
        self.total_latency = totlat_l + dist_l*size_l + totlat_h + dist_h*size_h
        self.total_risk = totrisk_l + totrisk_h + self.weight*self.total_latency

    def findable_nodes(self):
        return (node for node in self if node.findable)

    def nonfindable_nodes(self):
        return (node for node in self if not node.findable)

    def update_costs(self, gusher_map=None, start=BASKET_LABEL):
        """Update distances, latencies and risks of this node's descendants. Should be called on root of tree."""
        def recurse(node, parent_latency, total_predecessor_weight):
            if node.parent:
                if gusher_map:
                    node.distance = gusher_map.distance(node.parent.name, node.name)
                node.latency = parent_latency + node.distance
                node.risk = node.parent.risk + total_predecessor_weight*node.distance
            else:
                # Latency of root node is distance between start (i.e. basket) and root node
                if gusher_map:
                    node.latency = gusher_map.distance(start, node.name)
                    node.total_latency += node.latency*node.size
                else:
                    node.latency = 0
                node.risk = 0

            if node.high:
                recurse(node.high, node.latency, total_predecessor_weight + node.weight)
            if node.low:
                recurse(node.low, node.latency, total_predecessor_weight + node.weight)

        recurse(self, 0, 0)

    def calc_tree_score(self, gusher_map=None, start=BASKET_LABEL):
        """Calculate and store the total latency and total risk of the tree rooted at this node."""
        self.update_costs(gusher_map, start)
        self.total_latency, self.total_risk = 0, 0
        for node in self.findable_nodes():
            self.total_latency += node.latency
            self.total_risk += node.risk

    def validate(self, gusher_map=None):
        """Check that tree is a valid strategy tree."""
        def recurse(node, predecessors, possible_nodes):
            # can't open the same gusher twice
            if node.name in predecessors:
                raise ValidationError(node, f'gusher {node} already in set of opened gushers: {predecessors}')

            if possible_nodes:
                if node.name in possible_nodes:
                    possible_nodes.remove(node.name)
                    if not node.findable:
                        raise ValidationError(node, f'gusher {node} is incorrectly marked non-findable, '
                                                    f'should be {node.name}')
                elif node.name not in possible_nodes and node.findable:
                    raise ValidationError(node, f'gusher {node} is incorrectly marked findable, '
                                                f'should be {node.name + NEVER_FIND_FLAG}')

            if node.high or node.low:
                if not possible_nodes:
                    raise ValidationError(node, f'Goldie should have been found after opening gusher {node}')
                pred_new = predecessors.union({node.name})
                if gusher_map:
                    neighborhood = set(gusher_map.adj(node.name))
                else:
                    neighborhood = set()

                # make sure parent/child references are consistent
                if node.high:
                    assert node.high.parent is node, f'node = {node}, node.high = {node.high}, ' \
                                                     f'node.high.parent = {node.high.parent}'
                    recurse(node.high, pred_new, possible_nodes.intersection(neighborhood))
                if node.low:
                    assert node.low.parent is node, f'node = {node}, node.low = {node.low}, ' \
                                                    f'node.low.parent = {node.low.parent}'
                    recurse(node.low, pred_new, possible_nodes.difference(neighborhood))
            else:
                # reaching a leaf node must guarantee that the Goldie will be found
                if possible_nodes:
                    raise ValidationError(node, f'Goldie could still be in gushers {possible_nodes} '
                                                f'after opening gusher {node}')

        if gusher_map:
            unaccounted = set(gusher_map).difference(node.name for node in self.findable_nodes())
            if unaccounted:
                raise ValidationError(self, 'Strategy is not guaranteed to find Goldie if hiding in gushers ' +
                                            ', '.join(unaccounted))
        recurse(self, set(), set(gusher_map) if gusher_map else set())

    def get_costs(self, gusher_map=None):
        self.update_costs(gusher_map)
        latencies = {str(node): node.latency for node in self.findable_nodes()}
        risks = {str(node): node.risk for node in self.findable_nodes()}
        return latencies, risks

    def report(self, gusher_map=None, quiet=0):
        short_str = write_tree(self)
        long_str = write_instructions(self) + '\n'

        latencies, risks = self.get_costs(gusher_map)
        cost_long = f"times: {{{', '.join(f'{node}: {time:0.2f}' for node, time in sorted(latencies.items()))}}}\n"\
                    f"risks: {{{', '.join(f'{node}: {risk:0.2f}' for node, risk in sorted(risks.items()))}}}\n"
        cost_short = f"avg. time: {mean(latencies.values()):0.2f} +/- {pstdev(latencies.values()):0.2f}\n"\
                     f"avg. risk: {mean(risks.values()):0.2f} +/- {pstdev(risks.values()):0.2f}"

        output = short_str
        if quiet < 3:
            output = '-'*len(short_str) + '\n' + output + '\n'
            if quiet < 2:
                output += long_str + cost_long
            output += cost_short
        return output

    def get_adj_dict(self):
        adj_dict = {str(node): dict() for node in self}
        for node in self:
            if node.parent:
                depth = adj_dict[str(node.parent)][str(node)]['depth']
            else:
                depth = 1
            if node.high and node.low:
                children = {str(node.high): {'depth': depth*2, 'high': 1},
                            str(node.low): {'depth': depth*2, 'high': 0}}
            elif node.high:
                children = {str(node.high): {'depth': depth*2, 'high': 1}}
            elif node.low:
                children = {str(node.low): {'depth': depth*2, 'high': 0}}
            else:
                children = {}
            adj_dict[str(node)].update(children)

        return adj_dict


# Custom exception for invalid strategy trees
class ValidationError(Exception):
    def __init__(self, node, message):
        super().__init__(node, message)


def write_tree(root):
    """Write the strategy encoded by the subtree rooted at 'root' in modified Newick format.
    V(H, L) represents the tree with root node V, high subtree H, and low subtree L.
    A node name followed by * indicates that the gusher is being opened solely for information and the Goldie will
    never be found there."""
    if root.high and root.low:
        return f'{root}({write_tree(root.high)}, {write_tree(root.low)})'
    elif root.high:
        return f'{root}({write_tree(root.high)},)'
    elif root.low:
        return f'{root}(,{write_tree(root.low)})'
    else:
        return f'{root}'


# Strategy tree grammar
node = Regex(rf'\w+[{NEVER_FIND_FLAG}]?')
LPAREN, COMMA, RPAREN = map(Suppress, '(,)')
tree = Forward()
subtree = Group(Optional(tree))
subtrees = LPAREN - subtree.setResultsName('high') - COMMA - subtree.setResultsName('low') - RPAREN
tree << node.setResultsName('root') - Optional(subtrees)


def read_tree(tree_str, gusher_map, start=BASKET_LABEL):
    """Read the strategy encoded in tree_str and build the corresponding decision tree.
    V(H, L) represents the tree with root node V, high subtree H, and low subtree L.
    A node name followed by * indicates that the gusher is being opened solely for information and the Goldie will
    never be found there."""

    def build_tree(tokens):  # recursively convert ParseResults object into GusherNode tree
        findable = tokens.root[-1] is not NEVER_FIND_FLAG
        rootname = tokens.root.rstrip(NEVER_FIND_FLAG)
        try:
            root = GusherNode(rootname, gusher_map=gusher_map, findable=findable)
        except KeyError as err:
            raise ValueError(f"Couldn't find gusher {err}!") from None
        else:
            if tokens.high or tokens.low:
                high, low = None, None
                dist_h, dist_l = 1, 1
                if tokens.high:
                    high = build_tree(tokens.high)
                    try:
                        dist_h = gusher_map.distance(rootname, high.name)
                    except KeyError:
                        raise ValueError(f"No connection between {rootname} and {high.name}!") from None
                if tokens.low:
                    low = build_tree(tokens.low)
                    try:
                        dist_l = gusher_map.distance(rootname, low.name)
                    except KeyError:
                        raise ValueError(f"No connection between {rootname} and {low.name}!") from None
                root.add_children(high=high, low=low, dist_h=dist_h, dist_l=dist_l)
            return root

    tokens = tree.parseString(tree_str, parseAll=True)
    root = build_tree(tokens)
    root.calc_tree_score(gusher_map, start)
    return root


def write_instructions(tree):
    """Convert strategy tree into human-readable instructions."""
    def recurse(subtree, depth):
        indent = "   "*depth
        result = ""
        if subtree.size > 2 or (subtree.high and subtree.low):
            result += f"open {subtree}\n"
            if subtree.high:
                result += indent + f"{subtree} high --> " + recurse(subtree.high, depth+1)
            if subtree.low:
                result += indent + f"{subtree} low --> " + recurse(subtree.low, depth+1)
        else:
            result = ', '.join(str(node) for node in subtree) + '\n'
        return result
    return recurse(tree, 0).strip('\n ')



from .GusherMap import GusherMap, BASKET_LABEL
from .GusherNode import GusherNode, NEVER_FIND_FLAG, write_tree
from copy import deepcopy


def flag(findable):
    if findable:
        return ""
    else:
        return NEVER_FIND_FLAG


def get_strat_greedy(gusher_map):
    """Build a decision tree for the gusher gushers. Greedy algorithm not guaranteed to find the optimal tree,
    but should still return something decent."""
    def recurse(suspected):
        n = len(suspected)
        # Base cases
        if n == 0:
            return None
        if n == 1:
            return GusherNode(list(suspected.nodes)[0], gusher_map=gusher_map)

        # Choose vertex V w/ lowest penalty and degree closest to n/2
        min_weight = min(gusher_map.weight(v) for v in suspected)
        candidates = [v for v in suspected if gusher_map.weight(v) == min_weight]
        vertex = min(candidates, key=lambda v: abs(suspected.degree[v] - n/2))

        # Build subtrees
        suspect_if_high = set(gusher_map.adj(vertex))
        suspect_if_low = set(gusher_map).difference(suspect_if_high)
        suspect_if_low.remove(vertex)
        high = recurse(suspect_if_high)
        low = recurse(suspect_if_low)

        # Construct optimal tree
        root = GusherNode(vertex, gusher_map=gusher_map)
        root.add_children(high, low)
        return root

    return recurse(gusher_map.connections)


def get_strat(gushers, start=BASKET_LABEL, tuning=0.5, all_distances=None, all_weights=None, debug=False):
    """Build the optimal decision tree for a gusher map. Memoized algorithm."""
    def print_log(*args, **kwargs):
        if debug:
            print(*args, **kwargs)

    def distance(start, end):
        return all_distances if all_distances else gushers.distance(start, end)

    def weight(vertex):
        return all_weights if all_weights else gushers.weight(vertex)

    def score(latency, risk):
        return tuning*risk + (1-tuning)*latency

    def candidate_cost(candidate, latest_open):
        latency = candidate.total_latency + distance(latest_open, candidate.name)*candidate.size
        risk = candidate.total_risk + weight(latest_open)*latency
        return latency, risk

    solved_subgraphs = dict()
    # dict that associates a subgraph with its solution subtrees and their objective scores
    # stores deep copies of trees to avoid entangling references between different candidates

    def recurse(suspected, opened, solved):
        """Return the optimal subtree to follow given a set of suspected gushers, a set of opened gushers,
        and the most recently opened gusher."""
        # First 3 arguments refer to gushers using strings, but recurse() returns a GusherNode
        # suspected = set of unopened gushers that might have the Goldie
        # opened = tuple of opened gushers in the order they were opened

        # Base cases
        n = len(suspected)
        if n == 0:
            return None
        if n == 1:
            return GusherNode(list(suspected)[0], gushers)

        candidates = list()
        key = (frozenset(suspected), frozenset(opened))
        key_str = f'({", ".join(str(u) for u in suspected)} | {", ".join(f"~{o}" for o in opened)})'
        if key in solved:  # Don't recalculate subtrees for subgraphs we've already solved
            candidates = deepcopy(solved[key])
        else:
            # Generate best subtrees for this subgraph
            search_set = set(gushers).difference(opened)
            for vertex in search_set:
                findable = vertex in suspected
                neighborhood = set(gushers.adj(vertex))
                suspect_if_high = suspected.intersection(neighborhood)
                suspect_if_low = suspected.difference({vertex}).difference(neighborhood)
                # suspect_if_high, suspect_if_low = split(suspected.difference({vertex}), neighborhood)
                # Don't open non-suspected gushers that are adjacent to all/none of the suspected gushers
                # Opening them can neither find the Goldie nor provide additional information about the Goldie
                if not findable and not (suspect_if_high and suspect_if_low):
                    continue
                print_log(f'{key_str}; check gusher {vertex}{flag(findable)}\n'
                          f'    adj: {tuple(suspect_if_high)}\n'
                          f'    non-adj: {tuple(suspect_if_low)}')
                opened_new = opened + tuple(vertex)
                high = recurse(suspect_if_high, opened_new, solved)
                low = recurse(suspect_if_low, opened_new, solved)
                dist_h, dist_l = 1, 1
                if high:
                    dist_h = distance(vertex, high.name)
                if low:
                    dist_l = distance(vertex, low.name)
                root = GusherNode(vertex, gusher_map=gushers, findable=findable)
                root.add_children(high, low, dist_h, dist_l)
                candidates.append(root)
                print_log(f'subgraph: {key_str}\n'
                          f'    candidate solution: {write_tree(root)}\n'
                          f'    score: {score(root.total_latency, root.total_risk):g}\n')
            solved[key] = deepcopy(candidates)

        latest_open = opened[-1]
        root = min(candidates, key=lambda tree: score(*candidate_cost(tree, latest_open)))
        print_log(f'{key_str}; options: \n' +
                  '\n'.join(f'    ~{latest_open}--{distance(latest_open, tree.name):0.2f}--> ' +
                            f'{tree}({tree.high}, {tree.low}), ' +
                            f'raw score: {score(tree.total_latency, tree.total_risk):0.2f}, ' +
                            f'final score: {score(*candidate_cost(tree, latest_open)):0.2f}'
                            for tree in candidates) +
                  f'\n    choose gusher {root}: {write_tree(root)}')
        return root

    print_log(f"(U | ~O) means gushers in U could have Goldie, gushers in O have already been opened\n"
              f"------------------------------------------------------------------------------------")
    root = recurse(set(gushers), tuple(start), solved_subgraphs)
    root.update_costs(gushers, start=start)
    return root


# TODO - move to separate test file
if __name__ == '__main__':
    import cProfile
    G = GusherMap('lo')

    greedy = get_strat_greedy(G)
    strat = get_strat(G, debug=True)
    print(greedy.report(G))
    print(strat.report(G))

    def profile(n=1):
        for i in range(n):
            get_strat(G)

    cProfile.run('[profile(10)]', sort='cumulative')



from setuptools import setup, find_packages
import pathlib
import re

# Grab contents of README
# https://realpython.com/pypi-publish-python-package/#configuring-your-package
HERE = pathlib.Path(__file__).parent.resolve()
README = (HERE/'README.md').read_text()

# https://stackoverflow.com/a/41110107
INIT = (HERE/'goldieseeker/__init__.py').read_text()


def get_from_init(var):
    result = re.search(fr'{var}\s*=\s*[\'"]([^\'"]*)[\'"]', INIT)
    return result.group(1)


setup(
        name='goldieseeker',
        version=get_from_init('__version__'),
        description='A command-line tool for generating and evaluating Goldie Seeking strategies',
        long_description=README,
        long_description_content_type='text/markdown',
        author='redoxwarfare',
        url='https://github.com/redoxwarfare/goldieseeker',
        license='MIT',
        classifiers=[
                'Programming Language :: Python :: 3',
                'License :: OSI Approved :: MIT License',
                'Development Status :: 4 - Beta'
        ],
        packages=find_packages(),
        package_data={'goldieseeker': ['maps/*.csv', 'maps/*.txt', 'images/*.png']},
        include_package_data=True,
        python_requires='>=3.6',
        install_requires=[
                'click',
                'networkx',
                'matplotlib',
                'numpy',
                'scipy',
                'pyparsing'
        ],
        entry_points={
                'console_scripts': [
                        'gseek = goldieseeker.__main__:main'
                ]
        }
)
